/* zfil: display contents of file in hex and ascii format */
/* Copyright (c) 2004 Makka  */
/* This program is released under the GNU General Public License */

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>

int getLength(char *);
int getHex(char *);
int getDec(char *);
void fmt(char *, int , char *);
void fmt2(unsigned int, char *);
int dwidth;

int main(int argc, char *argv[]) 
{
  struct stat statbuf[sizeof(stat)];
  char inbuf[32];
  char fmtbuf[200];
  char workbuf[256];
  char * plen = NULL; 
  char * poffset = NULL;
  char * filename = NULL; 

  char * p1, * p2;
  int  decmode = 0;               /* offset and length are decimal */
  unsigned int offset = 0;        /* start at offset into file */
  unsigned length = 0;            /* length of data to display */
  unsigned int endpoint;          /* calculated end point of display */
  unsigned int dspx;              /* temp variable */

  int i;
  int fd;
  int xwidth;                     /* characters per line either 16 or 32 */
  int rqlen;                      /* number of bytes requested to read */ 
  int dsp;                        /* displacement to start printing chars */
  int c = 0;
  off_t rc1;

/* get options passed to program */
  while (1)
    {
      c = getopt(argc, argv, "do:l:" );
      if (c == -1) break;
      switch (c)
	{
	case 'd':
	  decmode = 1;
	  break;
	case 'l':
	  plen = optarg;
	  break;
	case 'o':
	  poffset = optarg;
	  break;
	default:
	  return;
	}
    }

  //if (decmode) printf("decimal mode\n");
  //else printf("hex mode\n");

/* get filename */
  if ((argc - optind) > 2) { printf("too many arguments\n"); return; } 
  if ((argc - optind) < 1) { printf("must specify file name\n"); return; }
  else filename = argv[optind];
 
/* stat the file to get info */
  c = stat(filename, statbuf);
  if ( c == -1 )
    {
      if (errno == EACCES) { printf("permission denied\n"); return; }
      else if (errno == ENOENT) { printf("bad file name\n"); return; }
      else { printf ("error: errno =  %d\n", errno); return; }
    }

  if (statbuf->st_mode & S_IFDIR)
    { 
      printf ("%s is a directory\n", filename);
      return;
    }

  if (poffset) 
    { 
      dspx = getLength( poffset ); 
      if ( dspx > 8) { printf("offset too great\n"); return; }
      if (decmode ) 
	offset = getDec( poffset);
      else 
	offset = getHex( poffset);
      if (offset == -1) { printf("offset invalid \n"); return; }
    }

  if (plen) 
    {
      dspx = getLength( plen );
      if (dspx > 8) { printf("length to great\n"); return; }
      if (decmode ) 
	length = getDec( plen);
      else
	length = getHex( plen);
      if (length == -1) { printf("length invalid \n"); return; }
    }
  else
    length = 256;

  if ( offset > statbuf->st_size ) 
    { printf("offset greater than file length\n"); return; }
      
  endpoint = offset + length;
  if ( endpoint > statbuf->st_size ) 
    {
      endpoint = statbuf->st_size;
      length = endpoint - offset;
    }

  //  printf("offset = %d, length = %d\n", offset, length);

/* calculate width of offset tag */
  dspx = offset + length;
  dwidth = 4;
  if (dspx < 0x10000) dwidth = 4;
  else if (dspx < 0x100000 ) dwidth = 5;
  else if (dspx < 0x1000000 ) dwidth = 6;
  else if (dspx < 0x10000000 ) dwidth = 7;
  else dwidth = 8;

  xwidth = 16;
/* construct header */
  printf("zfil: %s; file length = %x/%d;  start = %x/%d  length = %x/%d\n", 
	 filename, 
	 statbuf->st_size, statbuf->st_size, 
	 offset, offset,
	 length,length);

/* position file */

  fd = open(filename, O_RDONLY);

  rc1 = lseek(fd, offset, SEEK_SET);

  while ( offset < endpoint )
    {
      dspx = dwidth + 1 + xwidth*3 + (xwidth/4)*2;

      for (i = 0; i <  dspx ; i++) fmtbuf[i] = ' ';

      rqlen = xwidth - offset%xwidth;
      dsp = offset%xwidth;
      if ( rqlen > endpoint - offset ) rqlen = endpoint - offset;
      c = read(fd, inbuf, rqlen);
      fmt2( offset, fmtbuf);

      fmt(inbuf, c, workbuf);
      p1 = fmtbuf + dwidth + 2;
      p2 = fmtbuf + dwidth + 4 + (xwidth)/4 - 1 + xwidth*2;

      for (i = 0; i <  c ; i++) 
	{
	  *(p1 + 2*i + 2*dsp + (dsp+i)/4 ) = workbuf[2*i];
	  *(p1 + 2*i + 2*dsp + 1 + (dsp+i)/4 ) = workbuf[2*i+1];

	  if ( (inbuf[i] >= '!') && (inbuf[i] <= '~') )
	    *(p2 + i + dsp + (dsp+i)/8 ) = inbuf[i];
	  else
	    *(p2 + i + dsp + (dsp+i)/8) = '.';
	}
      *(p2 + i + dsp + (dsp+i)/8 ) = '\n';
      *(p2 + i + dsp + (dsp+i)/8 + 1) = '\0';
      fputs(fmtbuf, stdout);
      offset = offset + c;
    }

  c = close(fd) ;
  return 0;
}

/* 
 *  check hex number valid and get value
 */

int getHex(char * str)
{
  int i = 0, x = 0; 
  unsigned int value = 0;
  while ( *(str+i) != '\0' )
    {
      if ( ( *(str+i) >= '0' ) && ( *(str+i) <= '9' ) )
	   x = *(str+i) - '0';
      else if  ( ( *(str+i) >= 'a' ) && ( *(str+i) <= 'f' ) )
	   x = *(str+i) - 'a' + 10;
      else if 
	( ( *(str+i) >= 'A' ) && ( *(str+i) <= 'F' ) )
	   x = *(str+i) - 'A' + 10;
      else return -1;
      value = value*16 + x;
      i++;
    }
  return value;
}

int getDec(char * str)
{
  int i = 0, x = 0; 
  unsigned int value = 0;
  while ( *(str+i) != '\0' )
    {
      if ( ( *(str+i) >= '0' ) && ( *(str+i) <= '9' ) )
	   x = *(str+i) - '0';
      else return -1;
      value = value*10 + x;
      i++;
    }
  return value;
}

/* 
 *  get length of string
 */
int  getLength(char * str)
{
  int i = 0; 
  while ( *(str+i) != '\0' ) i++;
  return i;
}

/* 
 * format offset
 */
void fmt2(unsigned int offset, char * buf)
{
  int i = 0;
  unsigned int x, tmp;
  char j;
  int fmtlen = 0;
  for (i = 8; i > 0 ; i-- )
    {
      tmp = offset;
      tmp = tmp >> (i - 1)*4;
      tmp = tmp & 0xf;
      if (tmp <= 9) j = tmp + '0';
      else if (tmp <= 15) j = tmp - 10 + 'a';
      else printf("ERROR!!!!");
      if (i <= dwidth) buf[fmtlen++] = j;
    }
}

/* format string  */
void fmt(char * inbuf, int bytes, char * obuf)
{
  int tmp;
  int k = 0;
  int x;
  char j;
  for (k = 0; k < bytes; k++ )
    {
      tmp = inbuf[k];
      tmp = tmp >> 4;
      tmp = tmp & 0xf;

      if (tmp <= 9) j = tmp + '0';
      else if (tmp <= 15) j = tmp - 10 + 'a';
      else printf("ERROR!!!!");
      obuf[2*k] = j;
      
      tmp = inbuf[k];
      tmp = tmp & 0xf;
      if (tmp <= 9) j = tmp + '0';
      else if (tmp <= 15) j = tmp - 10 + 'a';
      else printf("ERROR!!!!");
      obuf[2*k+1] = j;
    }
}
